{  $Id: ncimap.pas 6207 2003-08-29 19:33:48Z mk $

   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this software; see the file gpl.txt. If not, write to the
   Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   Created on April, 26st 2003 by Markus Kaemmerer <mk@happyarts.de>

   This software is part of the OpenXP project (www.openxp.de).
}

{ Klasse TIMAP }

{$I xpdefine.inc}

unit ncimap;

interface

uses
  SysUtils, Classes, NetcallIndy, IdIMAP4, IdException;

type
  EIMAP                 = EIdException;       { Vorfahr }
  EIMAPLocal            = class(EIMAP);       { Allgemein }
  EIMAPInvalidParameter = class(EIMAPLocal);  { Ungueltiger Parameter }

type
  TMyIdIMAP4 = class(TIdIMAP4)
  public
    property IPVersion;
  end;

  TIMAP = class(TNetcallIndy)
  private
    IdIMAP              : TMyIdIMAP4;

    FHost               : string;       { Hostadresse }
    FPort               : integer;      { Portnummer }
    FIPv6               : boolean;
    FSSLMethod          : integer;
    FUseExplicitTLS     : boolean;

    FSupportUIDLs       : Boolean;   { Flag: Server supports UIDLs }
    FAvailableUIDLs     : TStringList; { NEW available mail UIDLs }

    procedure InitVars;

  protected
    FOnlyNew            : Boolean;   { nur neue Mails holen }
    FLeaveUnchanged     : Boolean;   { Status gelesen/ungelesen belassen }
    FDeleteMail         : Boolean;   { Mail nach dem Abholen lschen }
    FUser, FPassword    : string;    { Identifikation }
    FMailCount, FFirstUnseen: Integer;
    FTag: Integer;                   { last used Tag }

    procedure SetPort(i: integer);
    procedure SetSSLMethod(i: integer);
    procedure SetUseExplicitTLS(b: boolean);

    function GetNewMailCount: Integer;

  public
    { List of already retrieved mail UIDLs }
    UIDLs: TStringList;

    constructor Create;
    constructor CreateWithHost(const s: string);
    destructor Destroy; override;

    property Port:integer read FPort write SetPort;
    property IPv6:boolean read FIPv6 write FIPv6;
    property SSLMethod:integer read FSSLMethod write SetSSLMethod;
    property UseExplicitTLS:boolean read FUseExplicitTLS write SetUseExplicitTLS;

    property User: string read FUser write FUser;
    property Password: string read FPassword write FPassword;
    { Take only new mails into account, remap mail numbers (if UIDL support) -
      initialize before Stat }
    property OnlyNew: Boolean read FOnlyNew write FOnlyNew;
    { Leave seen/unseen status }
    property LeaveUnchanged: Boolean read FLeaveUnchanged write FLeaveUnchanged;
    { Delete mail after retrieve }
    property DeleteMail: Boolean read FDeleteMail write FDeleteMail;
    { Count of mails in mbox }
    property MailCount: Integer read FMailCount;
    { Count of NEW mails }
    property NewMailCount: Integer read GetNewMailCount;
    { Number of last read mail - 0 with UIDL support and OnlyNew}
    property FirstUnseen: Integer read FFirstUnseen;

    function MapUIDL(Nr: Integer): Integer;

    { Verbindung herstellen und Anmelden (Auto-Login) }
    function Connect: boolean;
    function Connected: boolean;

    { Abmelden }
    procedure Disconnect;

    { -------- IMAP-Zugriffe }

    // Initializes UIDL and statistical variables, should be called after Connect
    function Stat: boolean;
    // Empfaengt eine Nachricht
    function Retr(ID: Integer; List: TStringList; var InfoString: string): boolean;
    // Loescht die angegebene Nachricht
    function Dele(ID: Integer): boolean;
    // Nachrichten endgltig lschen
    function Expunge: Boolean;
  end;

implementation

uses
  xpglobal, typeform, ProgressOutput, ncresource,
  IdExplicitTLSClientServerBase, IdIOHandler, IdMessage, IdSSLOpenSSL,
  IdGlobal;

const
  DefaultIMAPPort       = 143;


constructor TIMAP.Create;
begin
  inherited Create;
  InitVars;
end;

constructor TIMAP.CreateWithHost(const s: string);
begin
  Create;
  FHost := s;
end;

procedure TIMAP.InitVars;
begin
  FHost:='';
  FPort:= DefaultIMAPPort;
  FUser:='';
  FPassword:='';

  FOnlyNew := True;
  FLeaveUnchanged := False;
  FDeleteMail := True;
  FFirstUnseen := 1;
  FTag := 1000;

  FSupportUIDLs := True;
  UIDLs:=TStringList.Create;
  FAvailableUIDLs:=TStringList.Create;

  IdIMAP := TMyIdIMAP4.Create;
  IdIMAP.ReadTimeout := DefaultTimeout;
  IdIMAP.OnConnected := OnConnected;
end;

destructor TIMAP.Destroy;
begin
  try
    Disconnect;
  except
  end;
  IdIMAP.Free;
  UIDLs.Free;
  FAvailableUIDLs.Free;
  inherited Destroy;
end;

function TIMAP.Connect: boolean;
begin
  // Authorisierung bei IMAP immer noetig
  if (FUser='') or (FPassword='') then
    raise EIMAPLocal.CreateFmt(res_connect3, ['Invalid account info']); // Anmeldung fehlgeschlagen

  IdIMAP.Host := FHost;
  IdIMAP.Port := FPort;
  if (IPv6) then
    IdIMAP.IPVersion := Id_IPv6
  else
    IdIMAP.IPVersion := Id_IPv4;

  IdIMAP.Username := FUser;
  IdIMAP.Password := FPassword;
  IdIMAP.AuthType := iatUserPass;

  Output(mcVerbose,res_connect1, [FHost]);

  IdIMAP.Connect;

  Output(mcInfo,res_connect4, [FHost]); // Verbunden

  Result:= true;
end;

procedure TIMAP.Disconnect;
begin
  if Connected then
  begin
    Output(mcInfo,res_disconnect);
    try
      IdIMAP.Disconnect;
    except
    end;
  end;
end;

function TIMAP.Connected: boolean;
begin
  Result := IdIMAP.Connected;
end;

function TIMAP.GetNewMailCount: Integer;
begin
  if (FSupportUIDLs and OnlyNew) then
    Result := FAvailableUIDLs.Count
  else
    Result := IdIMAP.MailBox.RecentMsgs;
end;

function TIMAP.Stat: Boolean;
var
  UIDL: string;
  ServerUIDLs: TStringList;
  Nr: integer;
begin
  Result := False;
  if Connected then
  begin
    IdIMAP.SelectMailbox('inbox');
    FMailCount := IdIMAP.MailBox.TotalMsgs;
    if (OnlyNew) then
    begin
      if (FSupportUIDLs) then
      begin
        ServerUIDLs := TStringList.Create;
        try
          for Nr := 1 to FMailCount do
          begin
            if (IdIMAP.GetUID(Nr, UIDL)) then
            begin
              ServerUIDLs.Add(UIDL);
              if (UIDLs.IndexOf(UIDL) = -1) then
                FAvailableUIDLs.AddObject(UIDL, Pointer(Nr));
            end;
          end;
          // we're done, now delete old UIDLs not available from server anymore
          // from UIDL list to prevent it from getting big
          Nr := 0;
          while Nr < UIDLs.Count do
          begin
            if ServerUIDLs.IndexOf(UIDLs[Nr]) = -1 then
              UIDLs.Delete(Nr)
            else
              Inc(Nr);
          end;
        finally
          ServerUIDLs.Free;
        end;
        FFirstUnseen := FMailCount - (FAvailableUIDLs.Count - 1)
      end
      else
        FFirstUnseen := FMailCount - (IdIMAP.MailBox.RecentMsgs - 1)
    end
    else
      FFirstUnseen := 1;
    if (FFirstUnseen <= 0) then
      FFirstUnseen := FMailCount+1;

    Result := True;
  end;
end;

function TIMAP.MapUIDL(Nr: Integer): Integer;
begin
  if (not FSupportUIDLs) or (not OnlyNew) then
    result:=Nr
  else
    result:=Int64(Pointer(FAvailableUIDLs.Objects[Nr - FFirstUnseen]));
end;

function TIMAP.Retr(ID: Integer; List: TStringList; var InfoString: string): boolean;
var
  IdMessage: TIdMessage;
  Flags: TIdMessageFlagsSet;
  RestoreMessageStatus: Boolean;
  FlagsOk: Boolean;
  IsNewMessage: Boolean;
  Nr: integer;
begin
  Result := False;
  InfoString := '';
  if Connected then
  begin
    ID := MapUIDL(ID);
    IdMessage := TIdMessage.Create;
    try
      FlagsOk := IdIMAP.RetrieveFlags(ID, Flags);
      RestoreMessageStatus := LeaveUnchanged and FlagsOk;
      if (FSupportUIDLs) then
        IsNewMessage := True
      else
        IsNewMessage := (not FlagsOk or (mfRecent in Flags));
      if (not OnlyNew or IsNewMessage) then
      begin
        if (IdIMAP.Retrieve(ID, IdMessage)) then
        begin
          InfoString := IdMessageToStringList (IdMessage, List);
          Result := True;
          if FSupportUIDLs then
          begin
            // mark UIDL as retrieved
            for Nr := 0 to FAvailableUIDLs.Count - 1 do
            begin
              if Int64(Pointer(FAvailableUIDLs.Objects[Nr])) = ID then
              begin
                if UIDLs.IndexOf(FAvailableUIDLs[Nr]) = -1 then
                  UIDLs.Add(FAvailableUIDLs[Nr]);
                break;
              end;
            end;
          end;
          if (RestoreMessageStatus) then
          begin
            Exclude (Flags, mfRecent);
            IdIMAP.StoreFlags([ID], sdReplace, Flags);
          end;
          if (DeleteMail) then
            Dele(ID);
        end
        else
          InfoString := res_msg_not_avail + #13#10;
      end
    finally
      IdMessage.Free;
    end;
  end;
end;

function TIMAP.Dele(ID: Integer): boolean;
begin
  Result := False;
  if Connected then
    Result := IdIMAP.DeleteMsgs([ID]);
end;

function TIMAP.Expunge: Boolean;
begin
  Result := False;
  if Connected then
    Result := IdIMAP.ExpungeMailBox;
end;

procedure TIMAP.SetPort(i: integer);
begin
  if (i >= 0) and (i <= 65535) then
    FPort:= i
  else
    raise EIMAPInvalidParameter.CreateFmt('Port must be in 0..65535 (%d)', [i]);
end;

procedure TIMAP.SetSSLMethod(i: integer);
var
  IOHandler: TIdIOHandler;
const
  SSLMethods: array[1..2] of TIdSSLVersion = (sslvSSLv3, sslvTLSv1_2);
begin
  IOHandler := IdIMAP.IOHandler;
  IdIMAP.IOHandler := nil;
  try
    case i of
      0: begin end;
      1,2:
        begin
          IdIMAP.IOHandler := TIdSSLIOHandlerSocketOpenSSL.Create(IdIMAP);
          TIdSSLIOHandlerSocketOpenSSL(IdIMAP.IOHandler).SSLOptions.Method := SSLMethods[i];
        end;
      else
        raise EIMAPInvalidParameter.CreateFmt('SSL-Method must be in 0=none/1=SSL/2=TLS (%d)', [i]);
    end;
  finally
    IOHandler.Free;
  end;
  SetUseExplicitTLS (FUseExplicitTLS);
end;

procedure TIMAP.SetUseExplicitTLS(b: boolean);
begin
  FUseExplicitTLS := b;
  if (IdIMAP.IOHandler <> nil) then
  begin
    if (FUseExplicitTLS) then
      IdIMAP.UseTLS := utUseExplicitTLS
    else
      IdIMAP.UseTLS := utUseImplicitTLS;
  end
  else
    IdIMAP.UseTLS := utNoTLSSupport;
end;

end.

