{  $Id: ncnntp.pas 6981 2005-08-21 07:41:55Z stell $

   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this software; see the file gpl.txt. If not, write to the
   Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   Created on July, 21st 2000 by Hinrich Donner <hd@tiro.de>

   This software is part of the OpenXP project (www.openxp.de).
}

{ Klasse TNNTP }

{$I xpdefine.inc}

unit ncnntp;

interface

uses
  SysUtils, Classes, NetcallIndy, IdNNTP, IdException;

{$I xpdefine.inc}

type
  ENNTP                 = EIdException;       { Vorfahr }
  ENNTPLocal            = class(ENNTP);       { Allgemein }
  ENNTPInvalidParameter = class(ENNTPLocal);  { Ungueltiger Parameter }
  ENNTP_Group_not_found = class(ENNTPLocal);

const
  nntp_NotConnected          = -1;

type
  TMyIdNNTP = class(TIdNNTP)
  public
    property IPVersion;
  end;

  TNNTP = class(TNetcallIndy)
  private
    IdNNTP              : TMyIdNNTP;

    FHost               : string;       { Hostadresse }
    FPort               : integer;      { Portnummer }
    FIPv6               : boolean;
    FSSLMethod          : integer;
    FUseExplicitTLS     : boolean;

    FUser, FPassword    : string;       { identification }
    FForceAuth          : boolean;

    FGroupsList         : TStringList;
    FGroupName          : String;

    procedure InitVars;

    procedure IdEvenTIdNewsgroupList (
      ANewsgroup: string; ALow, AHigh: Int64; AType: string; var ACanContinue: Boolean
    );

  protected
    procedure SetPort(i: integer);
    procedure SetSSLMethod(i: integer);
    procedure SetUseExplicitTLS(b: boolean);

  public
    constructor Create;
    constructor CreateWithHost(s: string);
    destructor Destroy; override;

    property Port:integer read FPort write SetPort;
    property IPv6:boolean read FIPv6 write FIPv6;
    property SSLMethod:integer read FSSLMethod write SetSSLMethod;
    property UseExplicitTLS:boolean read FUseExplicitTLS write SetUseExplicitTLS;

    property User: string read FUser write FUser;
    property Password: string read FPassword write FPassword;
    property ForceAuth: boolean read FForceAuth write FForceAuth;

    { Verbindung herstellen und Anmelden (Auto-Login) }
    function Connect: boolean;
    function Connected: boolean;

    { Abmelden }
    procedure Disconnect;

    { -------- NNTP-Zugriffe }

    { Liste holen }
    function List(aList: TStringList; onlyNew: Boolean; FromDateTime: TDateTime): boolean;

    { aktuelle Gruppe auswhlen }
    procedure SelectGroup(const AGroupName: String);

    { Message vom server holen }
    function GetMessage(Article, ArticleCount: Integer; Messages: TStringList; HeaderOnly: Boolean): Integer;
    function GetMessageByID(const MsgID: String; Messages: TStringList): Integer;

    { Message an server schicken }
    function PostMessage(Msg: TStringList): Integer;
    function PostPlainRFCMessages(Messages: TStringList): Integer;

    { next functions only available after selecting Group }

    { selected group }
    property GroupName: String read FGroupName;

    { first Message of this group hold by server }
    function FirstMessage: Integer;

    { last Message of this group hold by server }
    function LastMessage: Integer;

    { Posting allowed if ReadOnly = false }
    function ReadOnly: Boolean;
  end;


implementation

uses
  xpglobal, typeform, ProgressOutput, ncresource,
  IdExplicitTLSClientServerBase, IdSSLOpenSSL, IdIOHandler, IdMessage,
  IdGlobal;

const
  DefaultNNTPPort       = 119;

constructor TNNTP.Create;
begin
  inherited Create;
  InitVars;
end;

constructor TNNTP.CreateWithHost(s: string);
begin
  Create;
  FHost := s;
end;

procedure TNNTP.InitVars;
begin
  FHost:='';
  FPort:=DefaultNNTPPort;
  FUser:='';
  FPassword:='';

  IdNNTP := TMyIdNNTP.Create;
  IdNNTP.ReadTimeout := DefaultTimeout;
  IdNNTP.OnConnected := OnConnected;
  IdNNTP.OnNewsgroupList := IdEvenTIdNewsgroupList;
  IdNNTP.OnNewGroupsList := IdEvenTIdNewsgroupList;
end;

destructor TNNTP.Destroy;
begin
  try
    Disconnect;
  except
  end;
  IdNNTP.Free;
  inherited Destroy;
end;

function TNNTP.Connect: boolean;
begin
  IdNNTP.Host := FHost;
  IdNNTP.Port := FPort;
  if (IPv6) then
    IdNNTP.IPVersion := Id_IPv6
  else
    IdNNTP.IPVersion := Id_IPv4;

  IdNNTP.Username := FUser;
  IdNNTP.Password := FPassword;
  IdNNTP.ForceAuth := FForceAuth;

  Output(mcVerbose,res_connect1, [FHost]);

  IdNNTP.Connect;

  Output(mcInfo,res_connect4, [FHost]); // Verbunden

  Result := True
end;

procedure TNNTP.Disconnect;
begin
  if Connected then
  begin
    Output(mcInfo,res_disconnect);
    try
      IdNNTP.Disconnect;
    except
    end;
  end;
end;

function TNNTP.Connected: boolean;
begin
  Result := IdNNTP.Connected;
end;

procedure TNNTP.IdEvenTIdNewsgroupList (
  ANewsgroup: string; ALow, AHigh: Int64; AType: string; var ACanContinue: Boolean
);
begin
  FGroupsList.Add(ANewsgroup);
  if (FGroupsList.Count mod 25) = 0 then                        { User beruhigen }
    Output(mcVerbose, res_list4, [FGroupsList.Count]);
end;

function TNNTP.List(aList: TStringList; onlyNew: Boolean; FromDateTime: TDateTime): boolean;
begin
  Result := False;
  if Connected then
  begin
    aList.Duplicates:= dupIgnore;
    FGroupsList := aList;

    try
      { Liste anfordern }
      if OnlyNew and (not IsNAN(FromDateTime)) then
        IdNNTP.GetNewGroupsList(FromDateTime-OffsetFromUTC,True,'')
      else
        IdNNTP.GetNewsgroupList;

      Output(mcVerbose, res_list4, [aList.Count]);
    except
      Output(mcError,res_list2,[FHost]);
      raise;
    end;

    aList.Sort;  { Sortieren }

    Result := True;
  end;
end;

procedure TNNTP.SelectGroup(const AGroupName: String);
begin
  if Connected then
  try
    IdNNTP.SelectGroup(AGroupName);
    FGroupName := AGroupName;
  except
    Output(mcError, res_groupnotfound, [AGroupName]);
    raise;
  end;
end;

function TNNTP.GetMessage(Article, ArticleCount: Integer; Messages: TStringList; HeaderOnly: Boolean): Integer;
var
  i, iSize: Integer;
  Msg: TStringList;
  IdMessage: TIdMessage;
  GetArticleResult: Boolean;
begin
  Result := nntp_NotConnected;
  if Connected then
  begin
    Msg := TStringList.Create;
    IdMessage := TIdMessage.Create;
    try
      Result := 0;
      for Article := Article to Article+ArticleCount-1 do
      begin
        Msg.Clear;
        GetArticleResult := IdNNTP.GetHeader(Article, Msg);
        if (GetArticleResult and not HeaderOnly) then
        begin
          if (IdNNTP.GetArticle(Article, IdMessage)) then
            IdMessageToStringList (IdMessage, Msg)
          else
            GetArticleResult := False;
        end;

        if (not GetArticleResult) then
        begin
          Output(mcError, res_msg3, [IntToStr(Article)]);
          Continue;
        end;

        if (not HeaderOnly) then
          Inc(Result);

        // Lnge der Nachricht
        iSize := 0;
        for i := 0 to Msg.Count-1 do
          Inc (iSize, Length(Msg[i])+1);  // add one char for #1310 (see rfc so1036)

        Messages.Add('#! rnews ' + IntToStr(iSize));
        Messages.AddStrings(Msg);
        if (HeaderOnly) then
        begin
          Messages.Add('X-XP-MODE: HdrOnly'); // empty line to break up headers
          Messages.Add(''); // empty line to break up headers
        end;
      end;
    finally
      IdMessage.Free;
      Msg.Free;
    end;
  end;
end;

function TNNTP.GetMessageByID(const MsgID: String; Messages: TStringList): Integer;
var
  i, iSize: Integer;
  Msg: TStringList;
  IdMessage: TIdMessage;
  GetArticleResult: Boolean;
begin
  Result := nntp_NotConnected;
  if Connected then
  begin
    Msg := TStringList.Create;
    IdMessage := TIdMessage.Create;
    try
      Result := 0;
      GetArticleResult := IdNNTP.GetHeader(MsgID, Msg);
      if (GetArticleResult) then
      begin
        if (IdNNTP.GetArticle(MsgID, IdMessage)) then
        begin
          Msg.Add('');
          Msg.AddStrings(IdMessage.Body);
        end
        else
          GetArticleResult := False;
      end;

      if (not GetArticleResult) then
      begin
        Output(mcError, res_msg3, [MsgID]);
        Result := IdNNTP.LastCmdResult.NumericCode;
      end
      else
      begin
        // Lnge der Nachricht
        iSize := 0;
        for i := 0 to Msg.Count-1 do
          Inc (iSize, Length(Msg[i])+1);  // add one char for #1310 (see rfc so1036)
        Messages.Add('#! rnews ' + IntToStr(iSize));
        Messages.AddStrings(Msg);
        Result := 0;
      end;
    finally
      IdMessage.Free;
      Msg.Free;
    end;
  end;
end;

function TNNTP.PostMessage(Msg: TStringList): Integer;
var
  MessageStream: TStringStream;
  IdMessage: TIdMessage;
  NameValueSeparatorTmp: Char;
  From: string;
  posOparen, posCparen: integer;
const
  FromHeader: string = 'From';
begin
  Result := nntp_NotConnected;
  if Connected then
  begin
    Result := 0;
    MessageStream := TStringStream.Create('');
    IdMessage := TIdMessage.Create;
    NameValueSeparatorTmp := Msg.NameValueSeparator;
    try
      Msg.NameValueSeparator := ':';
      MessageStringsToStream (Msg, MessageStream);
      MessageStream.Position := 0;
      IdMessage.LoadFromStream(MessageStream);
      From := Msg.Values[FromHeader];
      if (From <> '') then
      begin
        posOparen := Pos('(',From);
        posCparen := Pos(')',From);
        if ((posOparen > 0) and (posCparen > 0) and (posCparen > posOparen)) then
        begin
          IdMessage.From.Address := Trim(Copy(From,1,posOparen-1));
          IdMessage.From.Name := Trim(Copy(From,posOparen+1,posCparen-posOparen-1));
        end;
      end;
      IdNNTP.Post(IdMessage);
      Result := 1;
    finally
      Msg.NameValueSeparator := NameValueSeparatorTmp;
      IdMessage.Free;
      MessageStream.Free;
    end;
  end;
end;

function TNNTP.PostPlainRFCMessages(Messages: TStringList): Integer;
var
  iLine, iPosting: Integer;
  Msg: TStringList;
  PostMessageResult: Integer;
begin
  Result := nntp_NotConnected;
  if Connected then
  begin
    Msg := TStringList.Create;
    try
      Result := 0;
      iLine := 0;
      iPosting := 0;
      while (iLine < Messages.Count) and (Pos('#! rnews', Messages[iLine]) = 1) do
      begin
        Inc(iLine);
        Inc(iPosting);
        Msg.Clear;
        while (iLine < Messages.Count) and (Pos('#! rnews', Messages[iLine]) <> 1) do
        begin
          Msg.Add(Messages[iLine]);
          Inc(iLine);
        end;
        Output(mcInfo,res_postmsg,[iPosting,(iLine/Messages.Count*100)]);
        PostMessageResult := PostMessage(Msg);
        if (PostMessageResult = nntp_NotConnected) then
          break
        else
          Inc(Result, PostMessageResult);
      end;
    finally
      Msg.Free;
    end;
  end;
end;

function TNNTP.FirstMessage: Integer;
begin
  Result := IdNNTP.MsgLow;
end;

function TNNTP.LastMessage: Integer;
begin
  Result := IdNNTP.MsgHigh;
end;

function TNNTP.ReadOnly: Boolean;
begin
  Result := (IdNNTP.Permission <> crCanPost);
end;

procedure TNNTP.SetPort(i: integer);
begin
  if (i >= 0) and (i <= 65535) then
    FPort:= i
  else
    raise ENNTPInvalidParameter.CreateFmt('Port must be in 0..65535 (%d)', [i]);
end;

procedure TNNTP.SetSSLMethod(i: integer);
var
  IOHandler: TIdIOHandler;
const
  SSLMethods: array[1..2] of TIdSSLVersion = (sslvSSLv3, sslvTLSv1_2);
begin
  IOHandler := IdNNTP.IOHandler;
  IdNNTP.IOHandler := nil;
  try
    case i of
      0: begin end;
      1,2:
        begin
          IdNNTP.IOHandler := TIdSSLIOHandlerSocketOpenSSL.Create(IdNNTP);
          TIdSSLIOHandlerSocketOpenSSL(IdNNTP.IOHandler).SSLOptions.Method := SSLMethods[i];
        end;
      else
        raise ENNTPInvalidParameter.CreateFmt('SSL-Method must be in 0=none/1=SSL/2=TLS (%d)', [i]);
    end;
  finally
    IOHandler.Free;
  end;
  SetUseExplicitTLS (FUseExplicitTLS);
end;

procedure TNNTP.SetUseExplicitTLS(b: boolean);
begin
  FUseExplicitTLS := b;
  if (IdNNTP.IOHandler <> nil) then
  begin
    if (FUseExplicitTLS) then
      IdNNTP.UseTLS := utUseExplicitTLS
    else
      IdNNTP.UseTLS := utUseImplicitTLS;
  end
  else
    IdNNTP.UseTLS := utNoTLSSupport;
end;


end.
