{  $Id: ncsmtp.pas 6981 2005-08-21 07:41:55Z stell $

   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this software; see the file gpl.txt. If not, write to the
   Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   Created on August, 15st 2000 by Markus Kmmerer <mk@happyarts.de>

   This software is part of the OpenXP project (www.openxp.de).
}

{ Klasse TSMTP }

{$I xpdefine.inc}

unit ncsmtp;

interface

uses
  SysUtils, Classes, NetcallIndy, IdSMTP, IdUserPassProvider, IdException;

type
  ESMTP                 = EIdException;       { Vorfahr }
  ESMTPLocal            = class(ESMTP);       { Allgemein }
  ESMTPInvalidParameter = class(ESMTPLocal);  { Ungueltiger Parameter }

type
  TMyIdSMTP = class(TIdSMTP)
  public
    property IPVersion;
  end;

  TSMTP = class(TNetcallIndy)
  private
    IdSMTP              : TMyIdSMTP;
    IdUserPassProvider  : TIdUserPassProvider;

    FHost               : string;       { Hostadresse }
    FPort               : integer;      { Portnummer }
    FIPv6               : boolean;
    FSSLMethod          : integer;
    FUseExplicitTLS     : boolean;

    FUser, FPassword    : string;       { Identifikation }
    FSecureLoginMandatory:Boolean;
    FromLine, ToLine    : Integer;

    procedure InitVars;

  protected
    procedure SetPort(i: integer);
    procedure SetSSLMethod(i: integer);
    procedure SetUseExplicitTLS(b: boolean);

    { Returns MailEndLine }
    { Parse complete SMTP-Mail }
    function ParseHeader(Mail: TStringList; StartLine: Integer;
                         Var From: String; EnvelopeTo: TStringList): Integer;

  public
    constructor Create;
    constructor CreateWithHost(s: string);
    destructor Destroy; override;

    property Port:integer read FPort write SetPort;
    property IPv6:boolean read FIPv6 write FIPv6;
    property SSLMethod:integer read FSSLMethod write SetSSLMethod;
    property UseExplicitTLS:boolean read FUseExplicitTLS write SetUseExplicitTLS;

    property User: string read FUser write FUser;
    property Password: string read FPassword write FPassword;
    property SecureLoginMandatory: boolean read FSecureLoginMandatory write FSecureLoginMandatory;

    { Verbindung herstellen und Anmelden (Auto-Login) }
    function Connect(AFQDomain: String): boolean;
    function Connected: boolean;

    { Abmelden }
    procedure Disconnect;

    { -------- SMTP-Zugriffe }

    procedure PostMail(
      Mail: TStringList;
      const EnvelopeFrom, Realname: string; EnvelopeTo: TStringList;
      var InfoString: string
    );
    procedure PostPlainRFCMails(Mail: TStringList; EnvelopeFrom, Realname: String);
    function GetFQDomain(Mail: TStringList): String;
  end;

implementation

uses
  xpglobal, typeform, ProgressOutput, ncresource,
  IdExplicitTLSClientServerBase, IdIOHandler,
  IdMessage, IdSSLOpenSSL, IdSASL_CRAM_MD5, IdSASLUserPass, IdGlobal;

const
  DefaultSMTPPort       = 25;

  SMTPFROMSIGN          = 'MAIL FROM:';
  SMTPTOSIGN            = 'RCPT TO:';
  SMTPDATASIGN          = 'DATA';
  SMTPHELOSIGN          = 'HELO';
  SMTPEHLOSIGN          = 'EHLO';
  SMTPQUITSIGN          = 'QUIT';


constructor TSMTP.Create;
begin
  inherited Create;
  InitVars;
end;

constructor TSMTP.CreateWithHost(s: string);
begin
  Create;
  FHost := s;
end;

procedure TSMTP.InitVars;
var
  IdSASLUserPass: TIdSASLUserPass;
begin
  FHost:='';
  FPort:=DefaultSMTPPort;
  FUser:='';
  FPassword:='';

  FromLine := -1;
  ToLine := -1;
  FSecureLoginMandatory := True;

  IdSMTP := TMyIdSMTP.Create;
  IdSMTP.ReadTimeout := DefaultTimeout;
  IdSMTP.OnConnected := OnConnected;

  IdUserPassProvider := TIdUserPassProvider.Create(IdSMTP);
  IdSASLUserPass := TIdSASLCramMD5.Create(IdSMTP);
  IdSASLUserPass.UserPassProvider := IdUserPassProvider;
  IdSMTP.SASLMechanisms.Add.SASL := IdSASLUserPass;
end;

destructor TSMTP.Destroy;
begin
  try
    Disconnect;
  except
  end;
  IdSMTP.Free;
  inherited Destroy;
end;

function TSMTP.Connect(AFQDomain: String): boolean;
begin
  // Falls AFQDomain = '' sorgt Indy fr eine passende Befllung
  IdSMTP.HeloName := AFQDomain;

  // Authorisierung
  IdSMTP.AuthType := satDefault;
  if ((FUser='') and (FPassword='')) then
    IdSMTP.AuthType  := satNone
  else if (FSecureLoginMandatory) then
  begin
    IdSMTP.AuthType := satSASL;
    IdUserPassProvider.Username := FUser;
    IdUserPassProvider.Password := FPassword;
  end;

  IdSMTP.Host := FHost;
  IdSMTP.Port := FPort;
  if (IPv6) then
    IdSMTP.IPVersion := Id_IPv6
  else
    IdSMTP.IPVersion := Id_IPv4;

  IdSMTP.Username := FUser;
  IdSMTP.Password := FPassword;

  Output(mcVerbose,res_connect1, [FHost]);

  IdSMTP.Connect;

  Output(mcInfo,res_connect4, [FHost]); // Verbunden

  Result:= true;
end;

procedure TSMTP.Disconnect;
begin
  if Connected then
  begin
    Output(mcInfo,res_disconnect);
    try
      IdSMTP.Disconnect;
    except
    end;
  end;
end;

function TSMTP.Connected: boolean;
begin
  Result := IdSMTP.Connected;
end;

procedure TSMTP.PostMail(
  Mail: TStringList;
  const EnvelopeFrom, Realname: string; EnvelopeTo: TStringList;
  var InfoString: string
);
var
  IdMessage: TIdMessage;
  MailStream: TStringStream;
  i: integer;
begin
  IdMessage := TIdMessage.Create;
  MailStream := TStringStream.Create('');
  try
    if FromLine = -1 then FromLine := 0;
    if ToLine = -1   then ToLine := Mail.Count - 1;

    for i := FromLine to ToLine do
    begin
      MailStream.WriteString(Mail[i]);
      MailStream.WriteString(EOL);
    end;
    MailStream.WriteString('.');
    MailStream.WriteString(EOL);

    MailStream.Position := 0;
    IdMessage.LoadFromStream(MailStream);
    IdMessage.Sender.Text := EnvelopeFrom;
    IdMessage.Sender.Name := Realname;
    IdMessage.From.Text := EnvelopeFrom;
    IdMessage.From.Name := Realname;
    IdMessage.Recipients.Clear;
    for i := 0 to EnvelopeTo.Count - 1 do
      IdMessage.Recipients.Add.Text := EnvelopeTo[i];

    InfoString := IdMessageToStringList (IdMessage);

    IDSMTP.Send(IdMessage);

  finally
    MailStream.Free;
    IdMessage.Free;
  end;
end;

{ Parses plain RFC mail buffer. Header should be from StartLine to StopLine.
  Returns end of mail line number. Stores mail body start and end line number
  in globals FromLine and ToLine. }
function TSMTP.ParseHeader(Mail: TStringList; StartLine: Integer;
                           Var From: String; EnvelopeTo: TStringList): Integer;
var
  I: Integer;
begin
  From := '';
  EnvelopeTo.Clear;
  ToLine := Mail.Count - 1;
  i := StartLine;
  while i < Mail.Count do
  begin
    if Pos(SMTPFROMSIGN, Mail[I]) = 1 then
      From := Copy(Mail[I], Length(SMTPFROMSIGN) + 1, Length(Mail[I]));
    if Pos(SMTPTOSIGN, Mail[I]) = 1 then
      EnvelopeTo.Add(Copy(Mail[I], Length(SMTPTOSIGN) + 1, Length(Mail[I])));
    if Pos(SMTPDATASIGN, Mail[I]) = 1 then
    begin
      FromLine := I + 1;
      break;
    end;
    if Pos(SMTPQUITSIGN, Mail[I]) = 1 then begin
      result := 0;
      exit;
    end;
    Inc(i);
  end;
  for I := i to Mail.Count - 1 do
  begin
    ToLine := I;
    if Mail[I] = '.' then break;
  end;
  Result := ToLine;
  Dec(ToLine);
end;

procedure TSMTP.PostPlainRFCMails(Mail: TStringList; EnvelopeFrom, Realname: String);
var
  From: String;
  iMail,currLine: Integer;
  EnvelopeTo: TStringList;
  InfoString: string;
  FirstMail: Boolean;
begin
  if Mail.Count < 4 then exit;
  EnvelopeTo := TStringList.Create;
  try
    currLine := 0;
    iMail := 1;
    FirstMail := True;
    while currLine < Mail.Count - 2 do
    begin
      currLine := ParseHeader(Mail, currLine, From, EnvelopeTo);
      if currLine = 0 then
        break;
      if (FirstMail) then
      begin
        FirstMail := False;
        Output(mcDefault,' ');
      end;
      Output(mcInfo, res_postmsg, [iMail, ((currLine + 2) / Mail.Count * 100)]);
      try
        InfoString := '';
        PostMail(Mail, iifs(EnvelopeFrom<>'',EnvelopeFrom,From), Realname, EnvelopeTo, InfoString);
      except
        if InfoString <> '' then
          Output(mcError, InfoString);
        raise;
      end;
      Output(mcDefault, InfoString);
      Inc(currLine); Inc(iMail);
    end;
  finally
    EnvelopeTo.Free;
  end;
end;

function TSMTP.GetFQDomain(Mail: TStringList): String;
begin
  if Mail.Count > 0 then
  begin
    if (Pos(SMTPHELOSIGN, Mail[0]) = 1) or (Pos(SMTPEHLOSIGN, Mail[0]) = 1) and (Pos(' ',Mail[0]) > 1) then
      Result := Copy(Mail[0], Pos(' ',Mail[0])+1, High(integer));
  end
  else
    Result := '';
end;

procedure TSMTP.SetPort(i: integer);
begin
  if (i >= 0) and (i <= 65535) then
    FPort:= i
  else
    raise ESMTPInvalidParameter.CreateFmt('Port must be in 0..65535 (%d)', [i]);
end;

procedure TSMTP.SetSSLMethod(i: integer);
var
  IOHandler: TIdIOHandler;
const
  SSLMethods: array[1..2] of TIdSSLVersion = (sslvSSLv3, sslvTLSv1_2);
begin
  IOHandler := IdSMTP.IOHandler;
  IdSMTP.IOHandler := nil;
  try
    case i of
      0: begin end;
      1,2:
        begin
          IdSMTP.IOHandler := TIdSSLIOHandlerSocketOpenSSL.Create(IdSMTP);
          TIdSSLIOHandlerSocketOpenSSL(IdSMTP.IOHandler).SSLOptions.Method := SSLMethods[i];
        end;
      else
        raise ESMTPInvalidParameter.CreateFmt('SSL-Method must be in 0=none/1=SSL/2=TLS (%d)', [i]);
    end;
  finally
    IOHandler.Free;
  end;
  SetUseExplicitTLS (FUseExplicitTLS);
end;

procedure TSMTP.SetUseExplicitTLS(b: boolean);
begin
  FUseExplicitTLS := b;
  if (IdSMTP.IOHandler <> nil) then
  begin
    if (FUseExplicitTLS) then
      IdSMTP.UseTLS := utUseExplicitTLS
    else
      IdSMTP.UseTLS := utUseImplicitTLS;
  end
  else
    IdSMTP.UseTLS := utNoTLSSupport;
end;

end.
